import asyncio

import aiohttp
from aiohttp import web

from v6d0auth.appfactory import AppFactory
from v6d0auth.run_app import start_app
from v6d1tokens.client import request_token
from v6d3music.api import Api
from v6d3music.core.set_config import set_config

__all__ = ("AppContext",)


class MusicAppFactory(AppFactory):
    def __init__(self, secret: str, api: Api):
        self.secret = secret
        self.loop = asyncio.get_running_loop()
        self._api = api

    def define_routes(self, routes: web.RouteTableDef) -> None:
        @routes.post("/api/")
        async def api(request: web.Request) -> web.Response:
            async with aiohttp.ClientSession() as s, s.get(
                "http://sessionservice/userid/", params={"session": request.query.get("session")}
            ) as response:
                user_id: int | None = await response.json()
            if user_id is None:
                raise web.HTTPUnauthorized
            if not isinstance(user_id, int):
                raise TypeError
            jsr = await request.json()
            assert isinstance(jsr, dict)
            try:
                return web.json_response(await self._api.api(jsr, user_id))
            except Api.MisusedApi as e:
                return web.json_response(e.json(), status=404)


class AppContext:
    def __init__(self, api: Api) -> None:
        self.api = api

    async def start(self) -> tuple[web.Application, asyncio.Task[None]] | None:
        try:
            factory = MusicAppFactory(await request_token("music-client", "token"), self.api)
            await set_config("secret", factory.secret)
        except aiohttp.ClientConnectorError:
            print("no web app (likely due to no token)")
        else:
            app = factory.app()
            task = asyncio.create_task(start_app(app))
            return app, task

    async def __aenter__(self) -> "AppContext":
        self.__task: asyncio.Task[tuple[web.Application, asyncio.Task[None]] | None] = asyncio.create_task(self.start())
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        if self.__task.done():
            result = await self.__task
            if result is not None:
                app, task = result
                await task
                await app.shutdown()
                await app.cleanup()
        else:
            self.__task.cancel()
