__all__ = ('Job', 'Pool', 'JobUnit', 'JobContext', 'JobStatusChanged', 'PoolEvent')

import asyncio
from typing import Any, Callable, Coroutine, Generic, TypeVar, Union

from v6d2ctx.integration.event import *
from v6d2ctx.integration.responsetype import *

from .abstractrunner import *


class JobEvent(Event):
    pass


class JobContext:
    def __init__(self, events: SendableEvents[JobEvent]) -> None:
        self.events = events


class Job:
    def __init__(self, future: asyncio.Future[None]) -> None:
        self.future = future

    async def run(self, context: JobContext, /) -> Union['Job', None]:
        raise NotImplementedError

    def json(self) -> ResponseType:
        return {'type': 'unknown'}


class JobUnit:
    async def run(self, context: JobContext, /) -> Union['JobUnit', None]:
        raise NotImplementedError

    def wrap(self) -> Job:
        return UnitJob(asyncio.Future(), self)

    def at(self, pool: 'Pool') -> 'JDC':
        return JDC(self, pool)

    def json(self) -> ResponseType:
        return {'type': 'unknown'}


class JobStatusChanged(JobEvent):
    def __init__(self, job: Job | JobUnit) -> None:
        self.job = job

    def json(self) -> ResponseType:
        return {'status': self.job.json()}


class UnitJob(Job):
    def __init__(self, future: asyncio.Future[None], unit: JobUnit) -> None:
        super().__init__(future)
        self.__unit = unit

    async def run(self, context: JobContext, /) -> Job | None:
        next_unit = await self.__unit.run(context)
        if next_unit is None:
            return None
        else:
            return UnitJob(self.future, next_unit)

    def json(self) -> ResponseType:
        return self.__unit.json()


class WorkerEvent(Event):
    pass


class _JWEvent(WorkerEvent):
    def __init__(self, event: JobEvent, /) -> None:
        self.event = event

    def json(self) -> ResponseType:
        return {'job': self.event.json()}


class _JWSendable(SendableEvents[JobEvent]):
    def __init__(self, sendable: SendableEvents[WorkerEvent]) -> None:
        self.sendable = sendable

    def send(self, event: JobEvent, /) -> None:
        return self.sendable.send(_JWEvent(event))


class Worker:
    def __init__(self, events: SendableEvents[WorkerEvent], /) -> None:
        self.__queue: asyncio.Queue[Job | None] = asyncio.Queue()
        self.__working = False
        self.__busy = 0
        self.__job = None
        self.__events = events
        self.__job_events = _JWSendable(self.__events)

    def _put_nowait(self, job: Job | None, /) -> None:
        self.__queue.put_nowait(job)
        self.__busy += 1

    async def _run(self, job: Job, /) -> Job | None:
        try:
            self.__job = job
            return await job.run(JobContext(self.__job_events))
        finally:
            self.__job = None

    async def _handle(self, job: Job) -> None:
        try:
            next_job = await self._run(job)
        except BaseException as e:
            job.future.set_exception(e)
        else:
            if next_job is None:
                job.future.set_result(None)
            else:
                self._put_nowait(next_job)

    async def _safe_handle(self, job: Job) -> None:
        try:
            await self._handle(job)
        except asyncio.InvalidStateError:
            pass

    def _on_shutdown(self) -> bool:
        if self.__queue.empty():
            self.__working = False
            return True
        else:
            self._put_nowait(None)
            return False

    async def _on(self, job: Job | None) -> bool:
        if job is None:
            return self._on_shutdown()
        else:
            await self._handle(job)
            return False

    async def _step(self) -> bool:
        job = await self.__queue.get()
        try:
            return await self._on(job)
        finally:
            self.__busy -= 1
            self.__queue.task_done()

    async def _work(self) -> None:
        while True:
            if await self._step():
                return

    def _cancel_job(self, job: Job) -> None:
        try:
            job.future.set_exception(RuntimeError('task left in the worker after shutdown'))
        except asyncio.InvalidStateError:
            pass

    def _cancel_one(self, job: Job | None) -> None:
        if job is not None:
            self._cancel_job(job)

    def _cancel_get(self) -> None:
        job = self.__queue.get_nowait()
        try:
            self._cancel_one(job)
        finally:
            self.__queue.task_done()

    def _cancel_all(self) -> None:
        while not self.__queue.empty():
            self._cancel_get()

    async def _task(self) -> None:
        await self._work()
        if self.__working:
            raise RuntimeError('worker left seemingly running after shutdown')
        if not self.__queue.empty():
            raise RuntimeError('worker failed to finish all its jobs')

    def start(self) -> asyncio.Future[None]:
        if self.__working:
            raise RuntimeError('starting an already running worker')
        self.__working = True
        return asyncio.create_task(self._task())

    def submit(self, job: Job | None) -> None:
        if not self.__working:
            raise RuntimeError('submitting to a non-working worker')
        self._put_nowait(job)

    def working(self) -> bool:
        return self.__working

    def busy(self) -> int:
        return self.__busy

    def _job_json(self) -> ResponseType:
        if self.__job is None:
            return None
        else:
            return self.__job.json()

    def json(self) -> ResponseType:
        return {
            'job': self._job_json(),
            'qsize': self.__queue.qsize(),
        }


class Working:
    def __init__(self, worker: Worker, task: asyncio.Future[None]) -> None:
        self.__worker = worker
        self.__task = task

    async def close(self) -> None:
        self.__worker.submit(None)
        await self.__task

    def submit(self, job: Job) -> None:
        self.__worker.submit(job)

    @classmethod
    def start(cls, events: SendableEvents[WorkerEvent], /) -> 'Working':
        worker = Worker(events)
        task = worker.start()
        return cls(worker, task)

    def working(self) -> bool:
        return self.__worker.working()

    def busy(self) -> int:
        return self.__worker.busy()

    def json(self) -> ResponseType:
        return self.__worker.json()


T = TypeVar('T')


class CoroJD(JobUnit, Generic[T]):
    def __init__(self, ctxcoro: Callable[[CoroContext], Coroutine[Any, Any, T]], /) -> None:
        self.future = asyncio.Future()
        self.ctxcoro = ctxcoro
        self.status: ResponseType = None

    async def run(self, context: JobContext, /) -> JobUnit | None:
        try:
            self.future.set_result(await self.ctxcoro(CoroContext(_CJSendable(context.events, self))))
        except BaseException as e:
            self.future.set_exception(e)
        return None

    def json(self) -> ResponseType:
        return {'coroutine': self.status}


class _CJSendable(SendableEvents[CoroEvent]):
    def __init__(self, sendable: SendableEvents[JobEvent], corojd: CoroJD) -> None:
        self.sendable = sendable
        self.corojd = corojd

    def send(self, event: CoroEvent, /) -> None:
        match event:
            case CoroStatusChanged() as csc:
                self.corojd.status = csc.status
                self.sendable.send(JobStatusChanged(self.corojd))
            case _:
                pass


class PoolEvent(Event):
    pass


class _WPEvent(PoolEvent):
    def __init__(self, event: WorkerEvent, /) -> None:
        self.event = event

    def json(self) -> ResponseType:
        return {'worker': self.event.json()}


class _WPSendable(SendableEvents[WorkerEvent]):
    def __init__(self, sendable: SendableEvents[PoolEvent], /) -> None:
        self.sendable = sendable

    def send(self, event: WorkerEvent, /) -> None:
        return self.sendable.send(_WPEvent(event))


class Pool(AbstractRunner):
    def __init__(self, workers: int, events: SendableEvents[PoolEvent], /) -> None:
        if workers < 1:
            raise ValueError('non-positive number of workers')
        self.__workers = workers
        self.__working = False
        self.__open = False
        self.__events: SendableEvents[PoolEvent] = events
        self.__worker_events: SendableEvents[WorkerEvent] = _WPSendable(self.__events)

    def _start_worker(self) -> Working:
        return Working.start(self.__worker_events)

    async def __aenter__(self) -> 'Pool':
        if self.__open:
            raise RuntimeError('starting an already open pool')
        if self.__working:
            raise RuntimeError('starting an already running pool')
        self.__working = True
        self.__pool: set[Working] = set(self._start_worker() for _ in range(self.__workers))
        self.__open = True
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        if not self.__working:
            raise RuntimeError('stopping a non-working pool')
        if not self.__open:
            raise RuntimeError('stopping a closed pool')
        self.__open = False
        for working in self.__pool:
            await working.close()
        del self.__pool
        self.__working = False

    def submit(self, job: Job) -> None:
        if not self.__working:
            raise RuntimeError('submitting to a non-working pool')
        if not self.__open:
            raise RuntimeError('submitting to a closed pool')
        min(self.__pool, key=lambda working: working.busy()).submit(job)

    def workers(self) -> int:
        return self.__workers

    async def run(self, coro: Coroutine[Any, Any, T], /) -> T:
        return await self.runctx(lambda _: coro)

    async def runctx(self, ctxcoro: Callable[[CoroContext], Coroutine[Any, Any, T]], /) -> T:
        job = CoroJD(ctxcoro)
        self.submit(job.wrap())
        return await job.future

    def json(self) -> ResponseType:
        return [working.json() for working in self.__pool]


class JDC:
    def __init__(self, unit: JobUnit, pool: Pool) -> None:
        self.__unit = unit
        self.__pool = pool

    async def __aenter__(self) -> 'JDC':
        job = self.__unit.wrap()
        self.__future = job.future
        self.__pool.submit(job)
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        await self.__future
